<?php
/**
 * @file
 *  Drupal Localize Fields module
 */

/**
 * The main reason for creating this class is that we need to pass info
 * from the hook_field_widget_form_alter() to the hook_field_form_alter() implementation.
 */
class LocalizeFields {

  /**
   * Variable 'localize_fields_usecontext' value when using translation context with no non-context fallback.
   *
   * @var integer
   */
  const USE_CONTEXT = 1;

  /**
   * Variable 'localize_fields_usecontext' value when using translation context and non-context as fallback.
   *
   * @var integer
   */
  const USE_CONTEXT_NONCONTEXT = 2;

  /**
   * Source language detection: use first enabled language as fallback instead of English.
   *
   * Would be daft, but that may actually happen (according to this maintainer's experience)
   * if no i18n_string_source_language variable and English isn't first language(?).
   *
   * @var boolean
   */
  const SOURCE_LANGUAGE_FALLBACK_FIRST_LANGUAGE = FALSE;

  /**
   * General context delimiter.
   *
   * @var string
   */
  const CONTEXT_DELIMITER = ':';

  /**
   * Context delimiter between bundle and field.
   *
   * @var string
   */
  const CONTEXT_DELIMITER_BUNDLE = '-';

  /**
   * Gets initialised by ::localize().
   *
   * @var integer
   */
  protected static $localize = -1;

  /**
   * Equivalent of conf var localize_fields_usecontext.
   *
   *  Values:
   *  0: no context
   *  1: context + fallback on no-context
   *  2: context only
   *
   * Gets initialised by ::localize().
   *
   * @var integer
   */
  protected static $useContext = -1;

  /**
   * @var string
   *
   * protected static $_userLanguage = 'en';
   */

  /**
   * @var array
   */
  protected static $entitiesRaw = array('"', "'");

  /**
   * Gets initialised by ::localize().
   *
   * @var array|NULL
   */
  protected static $entitiesEncoded;

  /**
   * @var array
   */
  protected static $formAlterFields = array();

  /**
   * Establish whether we should translate labels at all.
   *
   * Also works as (very light) init method, if deciding that that label translation should be done.
   *
   * @return integer
   *  Values 0|1; interprete as boolean.
   */
  public static function localize() {
    if (($localize = self::$localize) == -1) {
      $localize = 1;

      if (!empty($GLOBALS['language']->language)) {
        $userLanguage = $GLOBALS['language']->language;
      }
      // Definitely paranoid, but D7 language handling is tricky.
      elseif (!empty($GLOBALS['user']->language)) {
        $userLanguage = $GLOBALS['user']->language;
      }
      elseif (!($userLanguage = language_default('language'))) {
        $userLanguage = 'en';
      }

      // Establish source language by variable; fall back on English.
      if (!self::SOURCE_LANGUAGE_FALLBACK_FIRST_LANGUAGE) {
        $sourceLang = variable_get('i18n_string_source_language', 'en');
      }
      // Establish source language by variable; fall back on first enabled language (daft, but what appears to happen if no variable).
      elseif (!($sourceLang = variable_get('i18n_string_source_language'))) {
        $languages = language_list();
        $sourceLang = '';
        // Assume that source language has to be enabled(?).
        foreach ($languages as $langCode => $props) {
          if ($props->enabled) {
            $sourceLang = $langCode;
            break;
          }
        }
      }
      if (!$sourceLang || $userLanguage == $sourceLang) {
        $localize = 0;
      }
      else {
        self::$useContext = variable_get('localize_fields_usecontext', LocalizeFields::USE_CONTEXT_NONCONTEXT);
        self::$entitiesEncoded = variable_get('localize_fields_entsencoded', array('&quot;', '&#039;'));
      }
      self::$localize = $localize;
    }
    return $localize;
  }

  /**
   * Does NOT check for '' nor all-digits source - callers must do that.
   *
   * @param string $source
   * @param string $context
   * @param boolean $encoded
   * @return string
   */
  protected static function translateInternal($source, $context, $encoded = FALSE) {
    $useContext = self::$useContext;

    $translated = !$useContext ? t($source) : t($source, array(), array('context' => $context));

    if ($translated == $source) {
      if ($encoded && strpos($source, '&') !== FALSE) {
        // We do deliberately not use Drupal decode_entities() nor PHP html_entity_decode(),
        // because we want full control and only work on specific entities.

        // If the label actually was encoded...
        if (($decoded = str_replace(self::$entitiesEncoded, self::$entitiesRaw, $source)) != $source) {
          if (($translated_decoded = !$useContext ? t($decoded) : t($decoded, array(), array('context' => $context))) != $source) {
            // Re-encode after translation.
            return str_replace(self::$entitiesRaw, self::$entitiesEncoded, $translated_decoded);
          }
          // Fallback on no-context.
          elseif ($useContext == self::USE_CONTEXT_NONCONTEXT && ($translated_decoded = t($decoded)) != $source) {
            // Re-encode after translation.
            return str_replace(self::$entitiesRaw, self::$entitiesEncoded, $translated_decoded);
          }
        }
        elseif ($useContext == self::USE_CONTEXT_NONCONTEXT) { // ~ Fallback on no-context.
          return t($source);
        }
      }
      elseif ($useContext == self::USE_CONTEXT_NONCONTEXT) { // ~ Fallback on no-context.
        return t($source);
      }
    }

    return $translated;
  }

  /**
   * Translates a string taking possible translation context and possible encoding of single/double quotes into consideration.
   *
   * Utility method, for non-standard form manipulation.
   *
   * Check this module's help or settings page for context patterns, or use the ::context() method.
   * The context argument should not be empty if this module is set to use context + no non-context fallback (check settings page).
   *
   * @code
   * $label = LocalizeFields::translate(
   *   'whatever',
   *   array(),
   *   LocalizeFields::context('some_node_type', 'some_field_name', 'label')
   * );
   * @endcode
   *
   * @param string $source
   * @param array $args
   *  Ignored; args in field labels, description etc. aren't applicable (nor in core).
   * @param string $context
   *
   * @return string
   */
  public static function translate($source, $args = array(), $context = '') {
    if (($localize = self::$localize) == -1) { // Save a method call if possible.
      $localize = self::localize();
    }
    if (!$localize || $source === '' || ctype_digit('' . $source)) {
      return '' . $source;
    }
    // We need context if module set to use context + no non-context fallback.
    if (!$context && self::$useContext == self::USE_CONTEXT) {
      $ms = 'Needs non-empty context when module localize_fields\'s \'Use translation context\' setting is context + no non-context fallback';
      if (module_exists('inspect') && user_access('inspect log')) {
        inspect_trace(
          NULL,
          array(
            'category' => 'localize_fields',
            'message' => $ms,
            'severity' => WATCHDOG_WARNING,
          )
        );
      }
      else {
        watchdog(
          'localize_fields',
          __CLASS__ . '::' . __FUNCTION__ . ': ' . $ms . ', source[@source].',
          array('@source' => $source),
          WATCHDOG_WARNING
        );
      }

      // For what it's worth.
      return t($source);
    }

    // We don't want to bother folks with question about whether the source is encoded; let's just assume it is.
    return self::translateInternal($source, $context, TRUE);
  }

  /**
   * Get translation context.
   *
   * Utility method, normally not be needed.
   *
   * @param string $bundle
   *   Empty if field base property, like allowed_values.
   * @param string $field_name
   * @param string $property
   *
   * @return string
   *   Empty if non-contextual module setting.
   */
  public static function context($bundle, $field_name, $property) {
    if (self::$localize == -1) { // Save a method call if possible.
      self::localize();
    }
    if (self::$useContext) {
      $cntxtDelim = self::CONTEXT_DELIMITER;

      // Check for field base properties (allowed_values) to rule out stupid errors.
      if (!$bundle
        || $property == 'allowed_values'
      ) {
        return 'field' . $cntxtDelim . $field_name . $cntxtDelim . $property;
      }
      return 'field_instance' . $cntxtDelim . $bundle . self::CONTEXT_DELIMITER_BUNDLE . $field_name . $cntxtDelim . $property;
    }
    return '';
  }

  /**
   * Translates fields' labels, descriptions etc.
   *
   * Used by hook_field_widget_form_alter() implementation.
   *
   * Is executed before hook_form_alter().
   *
   * @param array $element
   * @param array $form_state
   * @param array $context
   */
  public static function fieldWidgetFormAlter(&$element, &$form_state, $context) {
    if (($localize = self::$localize) == -1) { // Save a method call if possible.
      $localize = self::localize();
    }
    if (!$localize || empty($context['field']['type'])) {
      return;
    }
    $cntxtDelim = self::CONTEXT_DELIMITER;

    $field_name = $context['field']['field_name'];
    $form_id = $form_state['build_info']['form_id'];

    $field_type = $context['field']['type'];

    $context_field = 'field' . $cntxtDelim . $field_name . $cntxtDelim;
    $context_instance = 'field_instance' . $cntxtDelim . $context['instance']['bundle'] . self::CONTEXT_DELIMITER_BUNDLE
      . $field_name . $cntxtDelim;

    // File fields are pretty non-standard.
    // And text/number field single cardinality may have any/multi cardinality structure.
    $file_cardinalitySingle = $structure_cardinalityNonSingle = FALSE;

    // These might be used more than once.
    $translated_label = $translated_description = '';

    // Find the array listing labels.
    if (array_key_exists('value', $element) && array_key_exists('#title', $element['value'])) {
      /*
       * text and number_ fields have a #title in a value bucket.
       *
       * Their single row instances' ['value']['#title'] is non-empty (and that one is used in widget rendering and validation),
       * whereas any/multiple rows instances' ['value']['#title'] is empty.
       *
       * Their single row instances also #title in in root; it's usually not used but let's play it safe
       *
       * text and number_ any/multiple rows instance is effectively first translated in the form_alter, because then they are suddenly non-empty.
       *
       * Single row instance:
       * #title: (string) >>Native title<<
       * value: (array) {
       * . #title: (string) >>Native title<<
       * }
       *
       * Any/multiple row instance:
       * value: (array) {
       * . #title: (string) >><<
       * }
       *
       * NB: sometimes single row instances have the same structure as any/multiple row instance.
       * Don't really know what affects/controls the structure.
       */
      $props =& $element['value'];

      // Sometimes single row instances have the same structure as any/multiple row instance.
      // Thus the instance wrapper MUST be translated (disregarding cardinality).
      $structure_cardinalityNonSingle = TRUE;

      // text and number_ fields single row instances got #title in root too; it's usually not used but let's play it safe.
      if (array_key_exists('#title', $element)) {
        if (($source = $element['#title']) !== '') {
          $element['#title'] = $translated_label = self::translateInternal($source, $context_instance . 'label', TRUE);
        }
        if (array_key_exists('#description', $element) && ($source = $element['#description']) !== '') {
          $element['#description'] = $translated_description = self::translateInternal($source, $context_instance . 'description', FALSE);
        }
      }
    }
    elseif (array_key_exists('#title', $element)) {
      // list_ types.
      $props =& $element;
    }
    elseif (array_key_exists(0, $element) && array_key_exists('#title', $element[0])) {
      // Single instance file has no props in $element root.
      $file_cardinalitySingle = TRUE;
      $props =& $element[0];
    }
    else {
      // Unsupported element structure.
      return;
    }

    // Field instances that allow any/more rows differ form equiv. single row instances, and typically have to be modified additionally in form_alter.
    $translateInstanceWrapper = $translateDescription = $preValidate = TRUE;

    // list type fields may need translation of settings.allowed_values (rarely used, but safer) in form_alter.
    $translateListOptionsSettings = FALSE;

    switch ($field_type) {
      case 'text':
      case 'text_long':
      case 'text_with_summary':
        if (!$structure_cardinalityNonSingle
          && !empty($context['field']['cardinality']) && $context['field']['cardinality'] == 1
        ) {
          // Doesn't need translation on instance wrapper level.
          $translateInstanceWrapper = FALSE;
        }
        break;
      case 'number_integer':
      case 'number_decimal':
      case 'number_float':
        if (!$structure_cardinalityNonSingle
          && !empty($context['field']['cardinality']) && $context['field']['cardinality'] == 1
        ) {
          // Doesn't need translation on instance wrapper level.
          $translateInstanceWrapper = FALSE;
        }
        // Prefix/suffixes are not encoded in form rendering.
        if (array_key_exists('#field_prefix', $props) && ($source = $props['#field_prefix']) !== '') {
          $props['#field_prefix'] = self::translateInternal($source, $context_instance . 'prefix', FALSE);
        }
        if (array_key_exists('#field_suffix', $props) && ($source = $props['#field_suffix']) !== '') {
          $props['#field_suffix'] = self::translateInternal($source, $context_instance . 'suffix', FALSE);
        }
        break;
      case 'list_boolean':
      case 'list_text':
      case 'list_integer':
      case 'list_decimal':
      case 'list_float':
        // Doesn't need translation on instance wrapper level.
        $translateInstanceWrapper = FALSE;
        // Translate option labels - unless flagged 'off'.
        if (empty($context['field']['settings']['allowed_values_no_localization'])
          && array_key_exists('#options', $props) && !empty($props['#options'])
        ) {
          // Needs translation of settings.allowed_values in form_alter.
          $translateListOptionsSettings = array();
          $anyOptionTranslation = FALSE;
          // Context is field name only, because optionss are a field_base property.
          $context_options = $context_field . 'allowed_values';

          foreach ($props['#options'] as $value => &$label) {
            // Option labels could easily be integers. Decimals on the other hand may have to be translated because of the decimal separator.
            if (!empty($label) && !ctype_digit('' . $label)) {
              $anyOptionTranslation = TRUE;
              // Option labels are not encoded in form rendering.
              $translateListOptionsSettings[$value] = $label = self::translateInternal($label, $context_options, FALSE);
            }
            else {
              // Flag no translation.
              $translateListOptionsSettings[$value] = NULL;
            }
          }
          unset($label); // Iteration ref.
          if (!$anyOptionTranslation) {
            $translateListOptionsSettings = FALSE;
          }
        }
        break;
      case 'file':
        // Doesn't need translation on instance wrapper level.
        // And we shan't prepend our own pre validator.
        $translateInstanceWrapper = $preValidate = FALSE;
        // Multi instance file fields have title and description on instances as well as the overall field.
        // And the title of such instances are used by core field validation.
        if (!$file_cardinalitySingle) {
          $limit = 100;
          for ($i = 0; $i < $limit; ++$i) {
            if (array_key_exists($i, $element)) {
              if (array_key_exists('#title', $element[$i]) && ($source = $element[$i]['#title']) !== '') {
                $element[$i]['#title'] = $translated_label !== '' ? $translated_label :
                  ($translated_label = self::translateInternal($source, $context_instance . 'label', TRUE));
              }
              if (array_key_exists('#description', $element[$i]) && ($source = $element[$i]['#description']) !== '') {
                $element[$i]['#description'] = $translated_description !== '' ? $translated_description :
                  ($translated_description = self::translateInternal($source, $context_instance . 'description', FALSE));
              }
            }
            else {
              break;
            }
          }
        }
        // Single instance (cardinality:1) file fields' #description
        // get file size and extensions info appended prior to this phase.
        elseif (array_key_exists('#description', $props) && ($source = $props['#description']) !== '' && strpos($source, '<br')) {
          $translateDescription = FALSE;

          $a = explode('<br', $source);
          $a[0] = t($a[0]);
          $props['#description'] = join('<br', $a);
        }
        break;
      case 'date':
      case 'datetime':
      case 'datestamp':
        // DOES need translation on instance wrapper level.

        // Add our custom date pre validator; because Date disregards hook_element_info_alter() implementations.
        // Otherwise localize_fields_element_info_alter() should add it.
        if (array_key_exists('#element_validate', $props)
          && !in_array('localize_fields_pre_element_validate', $props['#element_validate'])
        ) {
          array_unshift($props['#element_validate'], 'localize_fields_pre_element_validate');
        }
        break;
      // Default: Unknown field types may need translation on instance wrapper level.
    }

    // All types.
    if ($props['#title'] !== '') {
      $props['#title'] = $translated_label !== '' ? $translated_label :
        ($translated_label = self::translateInternal($props['#title'], $context_instance . 'label', TRUE));
    }

    // Most types.
    if ($translateDescription && array_key_exists('#description', $props) && $props['#description'] !== '') {
      $props['#description'] = $translated_description !== '' ? $translated_description :
        ($translated_description = self::translateInternal($props['#description'], $context_instance . 'description', FALSE));
    }

    if ($preValidate && array_key_exists('#element_validate', $props)
      && !in_array('localize_fields_pre_element_validate', $props['#element_validate'])
    ) {
      array_unshift($props['#element_validate'], 'localize_fields_pre_element_validate');
    }

    // If (non-file) multi-row core field type instance, or (all cardinalities) date/other non-core field type,
    // we have to change additionally via a form_alter, because the instance wrapper isnt passed to hook_field_widget_form_alter().
    // See field_multiple_value_form().
    if ($translateInstanceWrapper || $translateListOptionsSettings) {
      if (!empty($context['form']['#parents'])
        && !empty($context['form']['#language'])
      ) {
        $key_path = $context['form']['#parents'];
        $key_path[] = $field_name;
        $key_path[] = $context['form']['#language']; // Won't work if field has other (value) language than form.

        // Tell form_alter all about current field.
        $formAlters =& self::$formAlterFields;
        if (!$formAlters || !array_key_exists($form_id, $formAlters)) {
          $formAlters[$form_id] = array();
        }
        if ($translateInstanceWrapper) {
          $formAlters[$form_id][$context_instance] = array(
            'key_path' => $key_path,
            'label' => $translated_label,
            'description' => $translated_description,
          );
        }
        else { // $translateListOptionsSettings.
          $formAlters[$form_id][$context_field] = array(
            'key_path' => $key_path,
            'allowed_values' => $translateListOptionsSettings,
          );
        }
      }
    }
  }

  /**
   * Translates Date fields' 'Field-name Start date' and 'Field-name End date' labels.
   *
   * Used by hook_field_date_combo_process_alter() implementation.
   *
   * @param array &$element
   */
  public static function dateComboProcessAlter(&$element) {
    if (($localize = self::$localize) == -1) { // Save a method call if possible.
      $localize = self::localize();
    }
    if (!$localize) {
      return;
    }

    // The field label is already translated in another place - in the root of $element.
    $translated_label = $element['#title'];

    $items = array('value', 'value2');
    foreach ($items as $item) {
      if (array_key_exists($item, $element)) {
        // ['value']['#instance']['label']
        if (array_key_exists('#instance', $element[$item]) && array_key_exists('label', $element[$item]['#instance'])
          && ($source = $element[$item]['#instance']['label']) !== ''
        ) {
          // The label on the row is not translated yet.
          $element[$item]['#instance']['label'] = $translated_label;

          // ['value']['#date_title'] is label + 'start date'|'end date'
          if (array_key_exists('#date_title', $element[$item])
            && ($item_label = $element[$item]['#date_title']) !== '' && strpos($item_label, $source) !== FALSE
          ) {
            $element[$item]['#date_title'] = str_replace($source, $translated_label, $item_label);
          }
        }
      }
    }
  }

  /**
   * Translates instance wrapper label and description.
   *
   * Used by hook_field_form_alter() implementation.
   *
   * Is executed after hook_field_widget_form_alter().
   *
   * @param array &$form
   * @param array &$form_state
   * @param string $form_id
   */
  public static function formAlter(&$form, &$form_state, $form_id) {
    // No reason to check ::localize(), because ::field_widget_form_alter() only triggers this use if relevant.

    // Work on field instance wrappers or list type fields' options settings.
    if (self::$formAlterFields && array_key_exists($form_id, self::$formAlterFields)) {
      $fields =& self::$formAlterFields[$form_id];

      foreach ($fields as $context => $mem) {
        $key_path = $mem['key_path'];
        $field_name = $key_path[ count($key_path) - 2 ];
        $found = TRUE;
        $props =& $form;
        foreach ($key_path as $deeper_key) {
          if (array_key_exists($deeper_key, $props)) {
            $props =& $props[$deeper_key];
          }
          else {
            $found = FALSE;
            break;
          }
        }
        if ($found) {
          // list type fields' options settings.
          if (!empty($mem['allowed_values'])) {
            // Save in form_state; has no effect on values of field_info_field() but we can't do anything about that.
            $field_state = field_form_get_state($props['#field_parents'], $field_name, $props['#language'], $form_state);

            // Check for allowed_values_no_localization flag again - somebody may have set it after fieldWidgetFormAlter().
            if (empty($field_state['field']['settings']['allowed_values_no_localization'])
              && !empty($field_state['field']['settings']['allowed_values'])
            ) {
              $options =& $field_state['field']['settings']['allowed_values'];

              $context_options = $context . 'allowed_values';
              $anyOptionTranslation = FALSE;
              foreach ($options as $value => &$label) {
                if (array_key_exists($value, $mem['allowed_values'])) {
                  if (($translation = $mem['allowed_values'][$value]) !== NULL) {
                    $anyOptionTranslation = TRUE;
                    $label = $translation;
                  }
                }
                // Option labels could easily be integers. Decimals on the other hand may have to be translated because of the decimal separator.
                elseif ($value !== '' && !ctype_digit('' . $value)) {
                  $anyOptionTranslation = TRUE;
                  // Option labels are not encoded in form rendering.
                  $label = self::translateInternal($value, $context_options, FALSE);
                }
              }
              unset($label); // Iteration ref.
              unset($options); // Ref.

              if ($anyOptionTranslation) {
                field_form_set_state($props['#field_parents'], $field_name, $props['#language'], $form_state, $field_state);
              }
            }
          }
          // Field instance wrappers - label and description.
          else {
            if (array_key_exists('#title', $props) && ($source = $props['#title']) !== '') {
              // For text and number_ types the memorized translation _will_ be empty (because it was in field_widget_alter).
              $props['#title'] = $mem['label'] !== '' ? $mem['label'] :
                self::translateInternal($source, $context . 'label', TRUE);
            }
            if (array_key_exists('#description', $props) && ($source = $props['#description']) !== '') {
              $props['#description'] = $mem['description'] !== '' ? $mem['description'] :
                self::translateInternal($source, $context . 'description', FALSE);
            }
          }
        }
        unset($props); // Iteration ref.
      }

      // Releive memory.
      // And if the form_alter is (extraordinarily) invoked a second time during the request,
      // then labels shan't be translated again.
      unset($fields, self::$formAlterFields[$form_id]);
    }
  }

  /**
   * Custom element validator which performs no validation but translates the labels
   * used in element validation by field types having an #element_validate function.
   *
   *  Examples of #element_validate implementations:
   *  - number: number_field_widget_validate()
   *  - date: date_combo_validate()
   *
   * Does not translate the instance' description.
   *
   * For Date fields, Implementing hook_date_combo_pre_validate_alter() would _seem_ the right way to do this.
   * But date_combo_validate() fetches instance properties (like labels) _before_ invoking the date_combo_pre_validate_alter hook.
   * Thus changes in a hook_date_combo_pre_validate_alter() would have no effect.
   *
   * Attaching this validator is another challenge, Date disregards for instance implementations of hook_element_info_alter();
   * Date simply overrides (resets) the info in date_field_widget_form().
   * So this custom validator is attached (prepended) in our hook_field_widget_form_alter() implementation instead.
   *
   * @param array $element
   * @param array &$form_state
   */
  public static function preElementValidate($element, &$form_state) {
    if (($localize = self::$localize) == -1) { // Save a method call if possible.
      $localize = self::localize();
    }
    if (!$localize) {
      return;
    }

    // Non-standard field types may have weird structure - let's not break things further.
    if (empty($element['#field_parents']) || empty($element['#field_name']) || empty($element['#language'])) {
      return;
    }
    $field_name = $element['#field_name'];

    // Change the element by changing its' properties in $form_state.
    $field_state = field_form_get_state($element['#field_parents'], $field_name, $element['#language'], $form_state);

    if (array_key_exists('instance', $field_state) && array_key_exists('label', $field_state['instance'])
      && ($source = $field_state['instance']['label']) !== ''
    ) {
      $cntxtDelim = self::CONTEXT_DELIMITER;

      $field_state['instance']['label'] = self::translateInternal(
        $source,
        'field_instance' . $cntxtDelim . $element['#bundle'] . self::CONTEXT_DELIMITER_BUNDLE . $field_name . $cntxtDelim . 'label',
        TRUE
      );

      field_form_set_state($element['#field_parents'], $field_name, $element['#language'], $form_state, $field_state);
    }
  }

  /**
   * Translates error messages created by implementations of hook_field_validate().
   *
   * Also corrects decimal separator in decimals and floats.
   *
   * See https://drupal.org/node/1283718.
   *
   * Used by hook_field_attach_validate() implementation.
   *
   * @param array $entity
   * @param array &$errors
   */
  public static function fieldAttachValidate($entity, &$errors) {
    if (($localize = self::$localize) == -1) { // Save a method call if possible.
      $localize = self::localize();
    }
    if (!$localize) {
      return;
    }
    $cntxtDelim = self::CONTEXT_DELIMITER;

    $context_instance = '';
    // Find bundle.
    // $entity_type is useless here, because an entity type may pretty much any name.
    if (!empty($entity->form_id)) {
      // Bundle is form_id without '_form'.
      $context = 'field_instance' . $cntxtDelim . substr($entity->form_id, 0, strlen($entity->form_id) - 5) . self::CONTEXT_DELIMITER_BUNDLE;
    }
    elseif (!empty($entity->field_name)) {
      // Bundle is (field_collection) field_name.
      $context = 'field_instance' . $cntxtDelim . $entity->field_name . self::CONTEXT_DELIMITER_BUNDLE;
    }
    else {
      // Awful.
      if (module_exists('inspect') && user_access('inspect log')) {
        inspect(
          array(
            'entity' => $entity,
            'errors' => $errors,
          ),
          array(
            'depth' => 2,
            'category' => 'localize_fields',
            'message' => 'Can\'t find bundle equiv. property of entity',
            'severity' => WATCHDOG_WARNING,
          )
        );
      }
      else {
        watchdog(
          'localize_fields',
          __CLASS__ . '::' . __FUNCTION__ . ': Can\'t find bundle equiv. property of entity, erring field names: @field_names.',
          array('@field_names' => array_keys($errors)),
          WATCHDOG_WARNING
        );
      }

      return;
    }

    $lenPlaceholderStart = strlen($placeholderStart = '<em class="placeholder">');
    $lenPlaceholderEnd = strlen($placeholderEnd = '</em>');

    // $errors['some_field_name']['und'][0][0]['message']
    foreach ($errors as $field_name => $errors_byValueLanguage) {
      if ($context) {
        $context_instance = $context . $field_name . $cntxtDelim . 'label';
      }

      $decimal_separator = NULL;

      foreach ($errors_byValueLanguage as $language => $errors_byInstance) {
        foreach ($errors_byInstance as $instance => $errorList) {
          foreach ($errorList as $index => $error) {
            // error [
            //   'error': 'number_min',
            //   'message': '<em class="placeholder">The field label</em>: the value may be no less than <em class="placeholder">0.01</em>.'
            // ]
            if (!empty($error['message'])) {
              $message = $error['message'];
              // Do it thrice, because there may be more placeholders.
              for ($i = 0; $i < 3; ++$i) {
                if (($start = strpos($message, $placeholderStart)) !== FALSE && ($end = strpos($message, $placeholderEnd))) {
                  $start += $lenPlaceholderStart;
                  $msStart = str_replace($placeholderStart, '!PLACEHOLDER', substr($message, 0, $start));
                  $ms = substr($message, $start, $end - $start);
                  // Replace decimal separator...
                  if (is_numeric($ms)) {
                    if (strpos($ms, '.') !== FALSE) {
                      if (!$decimal_separator) {
                        $field_settings = field_info_field($field_name);
                        $decimal_separator = !empty($field_settings['settings']['decimal_separator']) ?
                          $field_settings['settings']['decimal_separator'] : '.';
                        unset($field_settings);
                      }
                      if ($decimal_separator != '.') {
                        $ms = str_replace('.', $decimal_separator, $ms);
                      }
                    }
                  }
                  // ...translate.
                  else {
                    $ms = self::translateInternal($ms, $context_instance, TRUE);
                  }
                  $message = $msStart
                    . $ms
                    . str_replace($placeholderEnd, '!_PLACEHOLDER', substr($message, $end, $lenPlaceholderEnd))
                    . substr($message, $end + $lenPlaceholderEnd);
                }
              }
              $message = str_replace(array('!PLACEHOLDER', '!_PLACEHOLDER'), array($placeholderStart, $placeholderEnd), $message);
              $errors[$field_name][$language][$instance][$index]['message'] = $message;
            }
          }
        }
      }
    }
  }

  /**
   * Translates field view labels, and corrects decimal separator in decimals/floats.
   *
   * Used by hook_field_attach_view_alter() implementation.
   *
   * @param array &$output
   */
  public static function fieldAttachViewAlter(&$output) {
    if (($localize = self::$localize) == -1) { // Save a method call if possible.
      $localize = self::localize();
    }
    if (!$localize) {
      return;
    }
    $cntxtDelim = self::CONTEXT_DELIMITER;

    // Find bundle.
    if (empty($output['#bundle'])) {
      // Abort if non-standard field type which misses the #bundle bucket.
      return;
    }
    $context = 'field_instance' . $cntxtDelim . $output['#bundle'] . self::CONTEXT_DELIMITER_BUNDLE;

    foreach ($output as $field_name => &$field) {
      if (strpos($field_name, 'field_') === 0 && is_array($field) && !empty($field['#field_type'])) {
        $context_instance = $context . $field_name . $cntxtDelim;
        if (!empty($field['#title'])) {
          $field['#title'] = self::translateInternal($field['#title'], $context_instance . 'label', TRUE);
        }
        switch ($field['#field_type']) {
          case 'list_boolean':
          case 'list_text':
          case 'list_integer':
          case 'list_decimal':
          case 'list_float':
            // Translate option labels.
            // Is a field_base property.
            $fieldInfo = field_info_field($field_name);
            $translateOptions = empty($fieldInfo['settings']['allowed_values_no_localization']);
            unset($fieldInfo);
            if ($translateOptions) {
              $context_options = 'field' . $cntxtDelim . $field_name . $cntxtDelim . 'allowed_values';
              $limit = 1000;
              for ($i = 0; $i < $limit; ++$i) {
                if (array_key_exists($i, $field)) {
                  // Options could easily be integers. Decimals on the other hand may have to be translated because of the decimal separator.
                  if (!empty($field[$i]['#markup']) && !ctype_digit($markup = '' . $field[$i]['#markup'])) {
                    // Option labels are not encoded when markup.
                    $field[$i]['#markup'] = self::translateInternal($markup, $context_options, FALSE);
                  }
                }
                else {
                  break; // Iter.
                }
              }
            }
            break; // Switch.
          case 'number_integer':
            // Translate prefix and/or suffix.
            $firstRow = TRUE;
            $prefixLength = $suffixLength = 0;
            $prefix = $suffix = '';
            $limit = 100;
            for ($i = 0; $i < $limit; ++$i) {
              if (array_key_exists($i, $field) && !empty($field[$i]['#markup'])) {
                if ($firstRow) {
                  $firstRow = FALSE;
                  if (!is_numeric($markup = $field[$i]['#markup'])) {
                    $instanceInfo = field_info_instance($output['#entity_type'], $field_name, $output['#bundle']);
                    if (($prefix = empty($instanceInfo['settings']['prefix']) ? '' : $instanceInfo['settings']['prefix'])) {
                      $prefixLength = strlen($prefix);
                      $prefix = self::translateInternal($prefix, $context_instance . 'prefix', FALSE);
                    }
                    if (($suffix = empty($instanceInfo['settings']['suffix']) ? '' : $instanceInfo['settings']['suffix'])) {
                      $suffixLength = strlen($suffix);
                      $suffix = self::translateInternal($suffix, $context_instance . 'suffix', FALSE);
                    }
                    unset($instanceInfo);
                    if ($prefixLength || $suffixLength) {
                      $field[$i]['#markup'] = $prefix . substr($markup, $prefixLength, strlen($markup) - ($prefixLength + $suffixLength)) . $suffix;
                    }
                    else {
                      // Don't work on any row. An error really, because the markup content should be numeric if no prefix or suffix.
                      break; // Iter.
                    }
                  }
                  else {
                    // No prefix or suffix: nothing to do.
                    break; // Iter.
                  }
                }
                // Won't get here unless there's a prefix or suffix.
                else {
                  $markup = $field[$i]['#markup'];
                  $field[$i]['#markup'] = $prefix . substr($markup, $prefixLength, strlen($markup) - ($prefixLength + $suffixLength)) . $suffix;
                }
              }
              else {
                // No more rows, or definitely no prefix/suffix.
                break; // Iter.
              }
            }
            break; // Switch.
          case 'number_decimal':
          case 'number_float':
            // Replace decimal separator if not dot.
            $fieldInfo = field_info_field($field_name);
            $decimal_separator = !empty($fieldInfo['settings']['decimal_separator']) ? $fieldInfo['settings']['decimal_separator'] : '.';
            unset($fieldInfo);
            // Translate prefix and/or suffix.
            $firstRow = TRUE;
            $prefixLength = $suffixLength = 0;
            $prefix = $suffix = '';
            $limit = 100;
            for ($i = 0; $i < $limit; ++$i) {
              if (array_key_exists($i, $field) && array_key_exists('#markup', $field[$i])) {
                if ($firstRow) {
                  $firstRow = FALSE;
                  if (!is_numeric($markup = $field[$i]['#markup'])) {
                    $instanceInfo = field_info_instance($output['#entity_type'], $field_name, $output['#bundle']);
                    if (($prefix = empty($instanceInfo['settings']['prefix']) ? '' : $instanceInfo['settings']['prefix'])) {
                      $prefixLength = strlen($prefix);
                      $prefix = self::translateInternal($prefix, $context_instance . 'prefix', FALSE);
                    }
                    if (($suffix = empty($instanceInfo['settings']['suffix']) ? '' : $instanceInfo['settings']['suffix'])) {
                      $suffixLength = strlen($suffix);
                      $suffix = self::translateInternal($suffix, $context_instance . 'suffix', FALSE);
                    }
                    unset($instanceInfo);
                    if ($prefixLength || $suffixLength) {
                      $value = substr($markup, $prefixLength, strlen($markup) - ($prefixLength + $suffixLength));
                      if ($decimal_separator != '.') {
                        $value = str_replace('.', $decimal_separator, $value);
                      }
                      $field[$i]['#markup'] = $prefix . $value . $suffix;
                    }
                    elseif ($decimal_separator != '.') {
                      $field[$i]['#markup'] = str_replace('.', $decimal_separator, $markup);
                    }
                    else {
                      // Nothing to do for any row.
                      break; // Iter.
                    }
                  }
                  elseif ($decimal_separator != '.') {
                    $field[$i]['#markup'] = str_replace('.', $decimal_separator, $markup);
                  }
                  else {
                    // Nothing to do for any row.
                    break; // Iter.
                  }
                }
                // Won't get here unless there's a prefix or suffix, or decimal separator must be fixed.
                else {
                  $markup = $field[$i]['#markup'];
                  if ($prefixLength || $suffixLength) {
                    $value = substr($markup, $prefixLength, strlen($markup) - ($prefixLength + $suffixLength));
                    if ($decimal_separator != '.') {
                      $value = str_replace('.', $decimal_separator, $value);
                    }
                    $field[$i]['#markup'] = $prefix . $value . $suffix;
                  }
                  else {
                    $field[$i]['#markup'] = str_replace('.', $decimal_separator, $markup);
                  }
                }
              }
              else {
                // No more rows, or a row misses #markup bucket (then an error).
                break; // Iter.
              }
            }
            break; // Switch.
        }
      }
    }
    unset($field); // Iteration ref
  }

}
